#include "saw.h"
#include "sawdoubler.h"
#include "lattice.h"
#include "symmetry.h"

void Iteration(int M, int N, int *visited, int *lattice, int *r, int *rl,
               struct treenode *Nodes, struct bin *Bins, struct vertex *V,
               int *Term, int nt, int rT, int *distT,
               int maxnum, nint *maxnodes, nint *max_nincludes,
               nint *niters, ccint *Z_2N, ccint *P_2N,
               int a, int b){

    /* This function performs one iteration of the main loop of the algorithm
       for walks with at least one site in the closed interval [a,b],
       no sites in the open interval (b, c), where c = Term[nt-1], 
       and ending in the termination set Term[nt-1], ... , Term[0].
       Just using the termination set can be achieved by setting a = -1.  */

    int S[MAXN] ;        /* set of sites S */
    nint nodes = 0 ;     /* number of nodes in the tree */
    int occursT ;        /* TRUE if site T=Term[0] occurs in the walk */
    cint nwalks ;        /* number of Type 1 walks created */
    cint nwalks2 ;       /* number of Type 2 walks created */
    struct gridpoint total ; /* total sum of the endpoint coordinates
                                and the squared endpoint distance for Type 1 walks */
    struct gridpoint total2 ; /* the same for Type 2 walks */
    int maxbin ;         /* maximum bin number of non-empty bin */
    nint nincludes = 0 ; /* number of includes of S so far */
    int countop ;        /* count operation, 1 if only walks of type 1, 
                                             2 if both type 1 and 2 */
    int EnoughMem ;      /* boolean: enough memory to store the tree */
    int i, symmS, c ;

    /* Copy Term into S */
    for (i=0; i<nt; i++)
        S[i] = Term[i] ;
    c = Term[nt-1] ;
    symmS = symmset(N, S, nt, lattice, rl) ;
    if (symmS==0)
        return ; 
    
    /* Count the iteration */
    (*niters)++ ;

    /* Determine count operation */
    if (M==0)
        countop = 1 ; /* walks of length N */
    else
        countop = 2 ; /* walks of length M and N */

    /* Initialise lattice visits and tree */
    r[0] = Origin(N) ;
    occursT = FALSE ;
    nwalks = 0 ;
    InitPoint(&total) ;
    for (i=0; i<Canon(N) ; i++)
        visited[i] = FALSE ;
    InitTree(Nodes, &nodes) ; 

    /* Create the walks of length N with Term as highest sites
       and add them to the tree */

    EnoughMem = TRUE ;
    go(0, N, N, 1, visited, lattice, r, &nwalks, &total, Nodes, &nodes,
       1, Term, nt, rT, distT, &occursT, &EnoughMem, a, b, c) ;
           
    if (M > 0){
        /* Also create walks of length M and add them to the same tree */
        r[0] = Origin(N) ;
        occursT = FALSE ;
        nwalks2 = 0 ;
        InitPoint(&total2) ;
        for (i=0; i<Canon(N) ; i++)
            visited[i] = FALSE ;

        go(0, M, N, 2, visited, lattice, r, &nwalks2, &total2, Nodes, &nodes,
           1, Term, nt, rT, distT, &occursT, &EnoughMem, a, b, c) ;
    }
        
    if (EnoughMem){
        struct zzout ZZout ; /* contains 2 integers, corrections for Z_2N and P_2N */

        if (nodes > *maxnodes)
            *maxnodes = nodes ;
                
        /* Initialise the bins */
        maxbin = maxnum ;
        InitBins(Nodes, nodes, Bins, &maxbin) ;
            
        ZZout = Correct (Nodes, N, countop, S, nt, DUMMY, &nincludes, 
                         Bins, maxbin, V, lattice, rl, a, b, c);
    
        if (nt%2 == 0){
            *Z_2N -= ZZout.Z ; 
            *P_2N -= ZZout.RR ; 
        } else {
            *Z_2N += ZZout.Z ; 
            *P_2N += ZZout.RR ; 
        }

        if (nincludes > *max_nincludes)
            *max_nincludes = nincludes ;

    } else {
        int a0, b0, a1, t, t1 ;

        /* Determine initial range */
        if (a == -1){
            a0 = 0 ;
            b0 = c -1 ;
        } else {
            a0 = a ;
            b0 = b ;
        }

        if (a0 <= b0){
            /* Find next splitting point t of interval
               using splitting ratio ALPHA, 0 < ALPHA < 1 */
            t = a0 + ALPHA * (b0-a0) ;
            if (t < a0) /* by rounding perhaps */
                t = a0 ;
            if (t > b0)
                t = b0 ; /* a0 <= t <= b0 */

            /* Reduce interval [a0,t] to [a0,a1] */
            a1 = t;
            while (a0 <= a1 && rl[a1] == DUMMY)
                a1-- ;
            while (a0 <= a1 && rl[a0] == DUMMY)
                a0++ ;
    
            /* Iteration for left interval */
            if (a0 == a1) {
                /* Include nondummy site a0 in terminal set */
                Term[nt] = a0 ;
                Iteration(M, N, visited, lattice, r, rl, Nodes, Bins, V,
                          Term, nt+1, rT, distT, maxnum, maxnodes,
                          max_nincludes, niters, Z_2N, P_2N, -1,-1) ;
            } else if (a0 < a1) {
                /* At least one site in [a0,a1] and none in (a1,c) */
                Iteration(M, N, visited, lattice, r, rl, Nodes, Bins, V,
                          Term, nt, rT, distT, maxnum, maxnodes,
                          max_nincludes, niters, Z_2N, P_2N, a0, a1) ;
    
            }

            /* Reduce interval [t+1,b0] to [t1,b0] */
            t1 = t+1;
            while (t1 <= b0 && rl[b0] == DUMMY)
                b0-- ;
            while (t1 <= b0 && rl[t1] == DUMMY)
                t1++ ;

            /* Iteration for right interval */
            if (t1 == b0) {
                /* Include nondummy site b0 in terminal set */
                Term[nt] = b0 ;
                Iteration(M, N, visited, lattice, r, rl, Nodes, Bins, V,
                          Term, nt+1, rT, distT, maxnum, maxnodes,
                          max_nincludes, niters, Z_2N, P_2N, -1,-1) ;
            } else if (t1 < b0) {
                /* At least one site in [t1,b0] and none in (b0,c) */
                Iteration(M, N, visited, lattice, r, rl, Nodes, Bins, V,
                          Term, nt, rT, distT, maxnum, maxnodes,
                          max_nincludes, niters, Z_2N, P_2N, t1, b0) ;
            }
        }
    }
    
    if (a == -1){ 
        ccint Z, RR ;

        /* Iteration is inclusive: add the terms for S[0..nt-1] */
        if (M==0){
            /* Compute Z and RR while avoiding overflow */
            Z = nwalks;
            Z *= nwalks ; /* Z = nwalks^2 */
            RR = nwalks ;
            RR *= total.rr ;
            RR -= Inprod(total, total) ;
            RR *= 2 ; /* RR = 2 * (nwalks*total.rr - total^2) */
        } else {
            ccint tmp1, tmp2 ;
            Z = nwalks ;
            Z *= nwalks2 ; /* Z = nwalks*nwalks2 */
            tmp1 = nwalks ;
            tmp1 *= total2.rr ;
            tmp2 = nwalks2 ;
            tmp2 *= total.rr ;
            RR = tmp1 + tmp2 - 2*Inprod(total, total2) ;
            /* RR = nwalks*total2.rr + nwalks2*total.rr - 2*Inprod(total, total2) */
        }

        Z *= symmS ;
        RR *= symmS ;
    
        if (nt%2 == 1) {
            Z = -Z ;     
            RR = -RR ;     
        }

        *Z_2N += Z ;
        *P_2N += RR ;
    }
        
} /* end Iteration */
