#include "saw.h"
#include "symmetry.h"

void InitPoint(struct gridpoint *point) {

    /* This function initialises point to 0 */

    point->x = 0 ;
    point->y = 0 ;
    point->z = 0 ;
    point->rr = 0 ;

} /* end InitPoint */


ccint Inprod(struct gridpoint point1, struct gridpoint point2) {

    /* This function computes the inner product of point1 and point2.
       It avoids overflow for large integers  by using
       longer temporary variables. */

    ccint tmpx, tmpy, tmpz ;

    tmpx = point1.x ;
    tmpx *= point2.x ;

    tmpy = point1.y ;
    tmpy *= point2.y ;

    tmpz = point1.z ;
    tmpz *= point2.z ;

    return tmpx + tmpy + tmpz ;

} /* end Inprod */


void InitNodeCounts(struct treenode *node) {

    /* This function initialises the counts of a treenode to 0 */

    node->Count = 0 ;
    node->Endpx = 0 ;
    node->Endpy = 0 ;
    node->Endpz = 0 ;
    node->Endprr = 0 ;
    node->Count2 = 0 ;
    node->Endpx2 = 0 ;
    node->Endpy2 = 0 ;
    node->Endpz2 = 0 ;
    node->Endprr2 = 0 ;

} /* end InitNodeCounts */


void AddCountsToNode(int ctype, struct treenode *node, struct gridpoint endpoint) {

    /* This function adds the counts of the endpoint to the node counts of Type ctype (1 or 2) */

    if (ctype == 1){
        node->Endpx += endpoint.x ;
        node->Endpy += endpoint.y ;
        node->Endpz += endpoint.z ;
        node->Endprr += endpoint.rr ;
    } else if (ctype == 2){
        node->Endpx2 += endpoint.x ;
        node->Endpy2 += endpoint.y ;
        node->Endpz2 += endpoint.z ;
        node->Endprr2 += endpoint.rr ;
    }

} /* end AddCounts */


void AddCountsToPoint(int ctype, struct treenode *node, struct gridpoint *endpoint) {

    /* This function adds the counts of the node of Type ctype (1 or 2) to the endpoint */

    if (ctype == 1){
        endpoint->x += node->Endpx ;
        endpoint->y += node->Endpy ;
        endpoint->z += node->Endpz ;
        endpoint->rr += node->Endprr ;
    } else if (ctype == 2){
        endpoint->x += node->Endpx2 ;
        endpoint->y += node->Endpy2 ;
        endpoint->z += node->Endpz2 ;
        endpoint->rr += node->Endprr2 ;
    }

} /* end AddCountsToPoint */


void AddNodeCounts(struct treenode *node, struct treenode *node2) {

    /* This function adds the counts of node2 to the counts of node */

        node->Count += node2->Count ;
        node->Count2 += node2->Count2 ;
        node->Endpx += node2->Endpx ;
        node->Endpy += node2->Endpy ;
        node->Endpz += node2->Endpz ;
        node->Endprr += node2->Endprr ;
        node->Endpx2 += node2->Endpx2 ;
        node->Endpy2 += node2->Endpy2 ;
        node->Endpz2 += node2->Endpz2 ;
        node->Endprr2 += node2->Endprr2 ;

} /* end AddNodeCounts */


void CopyNodeCounts(struct treenode *node, struct treenode *node2) {

    /* This function copies the counts of node2 to the counts of node */

        node->Count = node2->Count ;
        node->Count2 = node2->Count2 ;
        node->Endpx = node2->Endpx ;
        node->Endpy = node2->Endpy ;
        node->Endpz = node2->Endpz ;
        node->Endprr = node2->Endprr ;
        node->Endpx2 = node2->Endpx2 ;
        node->Endpy2 = node2->Endpy2 ;
        node->Endpz2 = node2->Endpz2 ;
        node->Endprr2 = node2->Endprr2 ;

} /* end CopyNodeCounts */


void CopyNodeCountsToVertex(struct treenode *node, struct vertex *v) {

    /* This function copies the counts from node to vertex v */

    v->Count  = node->Count ;
    v->Count2 = node->Count2 ;
    v->Endpx  = node->Endpx ;
    v->Endpy  = node->Endpy ;
    v->Endpz  = node->Endpz ;
    v->Endprr = node->Endprr ;
    v->Endpx2 = node->Endpx2 ;
    v->Endpy2 = node->Endpy2 ;
    v->Endpz2 = node->Endpz2 ;
    v->Endprr2 = node->Endprr2 ;

} /* end CopyNodeCountsToVertex */


void CopyNodeCountsFromVertex(struct treenode *node, struct vertex *v) {

    /* This function copies the counts from vertex v to node */

    node->Count  = v->Count ;
    node->Count2 = v->Count2 ;
    node->Endpx  = v->Endpx ;
    node->Endpy  = v->Endpy ;
    node->Endpz  = v->Endpz ;
    node->Endprr = v->Endprr ;
    node->Endpx2 = v->Endpx2 ;
    node->Endpy2 = v->Endpy2 ;
    node->Endpz2 = v->Endpz2 ;
    node->Endprr2 = v->Endprr2 ;

} /* end CopyNodeCountsFromVertex */


void reflect(int *i, int *j, int *k, int r) {

    /* This function executes the reflection r on lattice point (i, j, k),
       where -N <= i, j, k <= N and 0 <= r <= 7. Let r = (b2b1b0) in binary. */

       unsigned int one=1, two=2, four=4, ur ;

       ur = r ; /* unsigned r */

       if (ur&one)  *i = -(*i) ; /* b0 = 1 */
       if (ur&two)  *j = -(*j) ; /* b1 = 1 */
       if (ur&four) *k = -(*k) ; /* b2 = 1 */

} /* end reflect */


void permute(int *i, int *j, int *k, int r) {

    /* This function executes the permutation r on lattice point (i, j, k),
       where -N <= i, j, k <= N and 0 <= r <= 5. */

       int tmp ;

       switch(r){
           case 0: /* identity permutation */
               break ;
           case 1: /* swap i, j */
               tmp = *j ; *j = *i ; *i = tmp ;
               break;
           case 2: /* swap j, k */
               tmp = *j ; *j = *k ; *k = tmp ;
               break;
           case 3: /* rotate left */
               tmp = *i ; *i = *j ; *j = *k ; *k = tmp ;
               break ;
           case 4: /* rotate right */
               tmp = *k ; *k = *j ; *j = *i ; *i = tmp ; 
               break;
           case 5: /* swap i, k */
               tmp = *i ; *i = *k ; *k = tmp ;
               break ;
           default:
               printf("Internal error in permute\n");
               exit (1) ;
       }

       return ;

} /* end permute */
