#ifndef __symmetry_h__
#define __symmetry_h__

#include "saw.h"

/* Lattice dependent constants for 3D lattice */
#define SYMM 48 /* number of symmetries of the lattice */
#define REFL 8  /* number of reflection operations in 3D */

/* a and b are equivalent by one of the lattice symmetries
   if 48*t <= a, b < 48*(t+1), for some t.  */
#define Equiv(a, b)       ( ((a)/(SYMM)) == ((b)/(SYMM)) )

/* Origin (0,0,0) in the canonical numbering */
#define Origin(N)       ( (N)*(2*(N)+1)*(2*(N)+1) + (N)*(2*(N)+1) + (N) )

/* Structure for storing the (sum of)  coordinates of a gridpoint,
   with an extra entry for storing a squared radius */
typedef struct gridpoint {
    cint x ;   /* first coordinate */
    cint y ;   /* second coordinate  */
    cint z ;   /* third coordinate  */
    cint rr ;  /* squared radius 9or sum of radii), need not equal x^2+y^2+z^2 */
} Gridpoint ;

typedef struct treenode {
    int Site ;     /* site number, 0 <= site < MAXSITES */
    cint Count ;   /* count of number of walks of Type 1 ending in this node */
    cint Count2 ;  /* count of number of walks of Type 2 ending in this node */
    cint Endpx ;   /* sum of x coordinates of end points of walks of Type 1 ending in this node */
    cint Endpy ;   /* sum of y coordinates */
    cint Endpz ;   /* sum of z coordinates */
    cint Endprr ;  /* sum of squared Euclidean norms */
    cint Endpx2 ;   /* sum of x coordinates of end points of walks of Type 2 ending in this node */
    cint Endpy2 ;   /* sum of y coordinates */
    cint Endpz2 ;   /* sum of z coordinates */
    cint Endprr2 ;  /* sum of squared Euclidean norms */
    nint Child ;   /* first child, with lowest site number */
    nint Sibling ; /* in function go: next sibling, ordered by increasing site number
                      in function Correct: next node in linked list representing a bin */
    nint Parent ;  /* unique parent */
} Treenode ;

/* Similar structure, but without tree pointers */
typedef struct vertex {
    nint v ;       /* vertex number */
    cint Count ;   /* count of number of walks of Type 1 ending in this node */
    cint Count2 ;  /* count of number of walks of Type 2 ending in this node */
    cint Endpx ;   /* sum of x coordinates of end points of walks of Type 1 ending in this node */
    cint Endpy ;   /* sum of y coordinates */
    cint Endpz ;   /* sum of z coordinates */
    cint Endprr ;  /* sum of squared Euclidean norms */
    cint Endpx2 ;  /* sum of x coordinates of end points of walks of Type 2 ending in this node */
    cint Endpy2 ;  /* sum of y coordinates */
    cint Endpz2 ;  /* sum of z coordinates */
    cint Endprr2 ; /* sum of squared Euclidean norms */
} Vertex ;


/* Function declarations of symmetry functions */
void InitPoint(struct gridpoint *point) ;
ccint Inprod(struct gridpoint point1, struct gridpoint point2);
void InitNodeCounts(struct treenode *node) ;
void AddCountsToNode(int ctype, struct treenode *node, struct gridpoint endpoint) ;
void AddCountsToPoint(int ctype, struct treenode *node, struct gridpoint *endpoint) ;
void AddNodeCounts(struct treenode *node, struct treenode *node2) ;

void CopyNodeCounts(struct treenode *node, struct treenode *node2) ;
void CopyNodeCountsToVertex(struct treenode *node, struct vertex *v) ;
void CopyNodeCountsFromVertex(struct treenode *node, struct vertex *v) ;

int val(int i, int j, int k, int dir, int N, int *lattice) ;
int symmset(int N, int *S, int ns, int *lattice, int *rl) ;
void reflect(int *i, int *j, int *k, int r) ;
void permute(int *i, int *j, int *k, int r) ;
int Canon (int N) ;

#endif /* __symmetry_h__ */
